@@ -63,7 +63,11 @@ vec_proxy_equal.default <- function(x, ...) {
 vec_equal <- function(x, y, na_equal = FALSE, .ptype = NULL) {
   vec_assert(na_equal, ptype = logical(), size = 1L)
   args <- vec_recycle_common(x, y)
-  args <- vec_cast_common(!!!args, .to = .ptype)
+  args <- vec_cast_common_params(
+    !!!args,
+    .to = .ptype,
+    .df_fallback = DF_FALLBACK_QUIET
+  )
   .Call(vctrs_equal, args[[1]], args[[2]], na_equal)
 }
 
