@@ -70,7 +70,9 @@ vec_ptype_abbr.default <- function(x, ...) {
       character = "chr",
       complex = "cpl",
       list = "list",
-      expression = "expr"
+      expression = "expr",
+      raw = "raw",
+      abbreviate(typeof(x))
     )
     paste0(abbr, vec_ptype_shape(x))
   } else {
@@ -78,20 +80,6 @@ vec_ptype_abbr.default <- function(x, ...) {
   }
 }
 
-# AsIs --------------------------------------------------------------------
-
-#' @export
-vec_ptype_full.AsIs <- function(x, ...) {
-  class(x) <- setdiff(class(x), "AsIs")
-  paste0("I<", vec_ptype_full(x), ">")
-}
-
-#' @export
-vec_ptype_abbr.AsIs <- function(x, ...) {
-  class(x) <- setdiff(class(x), "AsIs")
-  paste0("I<", vec_ptype_abbr(x), ">")
-}
-
 # Helpers -----------------------------------------------------------------
 
 vec_ptype_shape <- function(x) {
