@@ -176,7 +176,7 @@ int compare_scalar(SEXP x, R_len_t i, SEXP y, R_len_t j, bool na_equal) {
 
 // -----------------------------------------------------------------------------
 
-static SEXP df_compare(SEXP x, SEXP y, bool na_equal, R_len_t n_row);
+static SEXP df_compare(SEXP x, SEXP y, bool na_equal, R_len_t size);
 
 #define COMPARE(CTYPE, CONST_DEREF, SCALAR_COMPARE)     \
 do {                                                    \
@@ -222,55 +222,43 @@ SEXP vctrs_compare(SEXP x, SEXP y, SEXP na_equal_) {
 
 // -----------------------------------------------------------------------------
 
-static struct vctrs_df_rowwise_info vec_compare_col(SEXP x,
-                                                    SEXP y,
-                                                    bool na_equal,
-                                                    struct vctrs_df_rowwise_info info,
-                                                    R_len_t n_row);
+static void vec_compare_col(int* p_out,
+                            struct df_short_circuit_info* p_info,
+                            SEXP x,
+                            SEXP y,
+                            bool na_equal);
 
-static struct vctrs_df_rowwise_info df_compare_impl(SEXP x,
-                                                    SEXP y,
-                                                    bool na_equal,
-                                                    struct vctrs_df_rowwise_info info,
-                                                    R_len_t n_row);
+static void df_compare_impl(int* p_out,
+                            struct df_short_circuit_info* p_info,
+                            SEXP x,
+                            SEXP y,
+                            bool na_equal);
 
-static struct vctrs_df_rowwise_info init_rowwise_compare_info(R_len_t n_row) {
-  struct vctrs_df_rowwise_info info;
+static SEXP df_compare(SEXP x, SEXP y, bool na_equal, R_len_t size) {
+  int nprot = 0;
+
+  SEXP out = PROTECT_N(Rf_allocVector(INTSXP, size), &nprot);
+  int* p_out = INTEGER(out);
 
   // Initialize to "equality" value
   // and only change if we learn that it differs
-  info.out = PROTECT(Rf_allocVector(INTSXP, n_row));
-  int* p_out = INTEGER(info.out);
-  memset(p_out, 0, n_row * sizeof(int));
+  memset(p_out, 0, size * sizeof(int));
 
-  // To begin with, no rows have a known comparison value
-  info.row_known = PROTECT(Rf_allocVector(RAWSXP, n_row * sizeof(bool)));
-  info.p_row_known = (bool*) RAW(info.row_known);
-  memset(info.p_row_known, false, n_row * sizeof(bool));
+  struct df_short_circuit_info info = new_df_short_circuit_info(size);
+  struct df_short_circuit_info* p_info = &info;
+  PROTECT_DF_SHORT_CIRCUIT_INFO(p_info, &nprot);
 
-  info.remaining = n_row;
-
-  UNPROTECT(2);
-  return info;
-}
-
-static SEXP df_compare(SEXP x, SEXP y, bool na_equal, R_len_t n_row) {
-  int nprot = 0;
-
-  struct vctrs_df_rowwise_info info = init_rowwise_compare_info(n_row);
-  PROTECT_DF_ROWWISE_INFO(&info, &nprot);
-
-  info = df_compare_impl(x, y, na_equal, info, n_row);
+  df_compare_impl(p_out, p_info, x, y, na_equal);
 
   UNPROTECT(nprot);
-  return info.out;
+  return out;
 }
 
-static struct vctrs_df_rowwise_info df_compare_impl(SEXP x,
-                                                    SEXP y,
-                                                    bool na_equal,
-                                                    struct vctrs_df_rowwise_info info,
-                                                    R_len_t n_row) {
+static void df_compare_impl(int* p_out,
+                            struct df_short_circuit_info* p_info,
+                            SEXP x,
+                            SEXP y,
+                            bool na_equal) {
   int n_col = Rf_length(x);
 
   if (n_col == 0) {
@@ -285,28 +273,24 @@ static struct vctrs_df_rowwise_info df_compare_impl(SEXP x,
     SEXP x_col = VECTOR_ELT(x, i);
     SEXP y_col = VECTOR_ELT(y, i);
 
-    info = vec_compare_col(x_col, y_col, na_equal, info, n_row);
+    vec_compare_col(p_out, p_info, x_col, y_col, na_equal);
 
     // If we know all comparison values, break
-    if (info.remaining == 0) {
+    if (p_info->remaining == 0) {
       break;
     }
   }
-
-  return info;
 }
 
 // -----------------------------------------------------------------------------
 
 #define COMPARE_COL(CTYPE, CONST_DEREF, SCALAR_COMPARE)              \
 do {                                                                 \
-  int* p_out = INTEGER(info.out);                                    \
-                                                                     \
   const CTYPE* p_x = CONST_DEREF(x);                                 \
   const CTYPE* p_y = CONST_DEREF(y);                                 \
                                                                      \
-  for (R_len_t i = 0; i < n_row; ++i, ++p_x, ++p_y) {                \
-    if (info.p_row_known[i]) {                                       \
+  for (R_len_t i = 0; i < p_info->size; ++i, ++p_x, ++p_y) {         \
+    if (p_info->p_row_known[i]) {                                    \
       continue;                                                      \
     }                                                                \
                                                                      \
@@ -314,30 +298,28 @@ do {                                                                 \
                                                                      \
     if (cmp != 0) {                                                  \
       p_out[i] = cmp;                                                \
-      info.p_row_known[i] = true;                                    \
-      --info.remaining;                                              \
+      p_info->p_row_known[i] = true;                                 \
+      --p_info->remaining;                                           \
                                                                      \
-      if (info.remaining == 0) {                                     \
+      if (p_info->remaining == 0) {                                  \
         break;                                                       \
       }                                                              \
     }                                                                \
   }                                                                  \
-                                                                     \
-  return info;                                                       \
 }                                                                    \
 while (0)
 
-static struct vctrs_df_rowwise_info vec_compare_col(SEXP x,
-                                                    SEXP y,
-                                                    bool na_equal,
-                                                    struct vctrs_df_rowwise_info info,
-                                                    R_len_t n_row) {
+static void vec_compare_col(int* p_out,
+                            struct df_short_circuit_info* p_info,
+                            SEXP x,
+                            SEXP y,
+                            bool na_equal) {
   switch (vec_proxy_typeof(x)) {
-  case vctrs_type_logical:   COMPARE_COL(int, LOGICAL_RO, lgl_compare_scalar);
-  case vctrs_type_integer:   COMPARE_COL(int, INTEGER_RO, int_compare_scalar);
-  case vctrs_type_double:    COMPARE_COL(double, REAL_RO, dbl_compare_scalar);
-  case vctrs_type_character: COMPARE_COL(SEXP, STRING_PTR_RO, chr_compare_scalar);
-  case vctrs_type_dataframe: return df_compare_impl(x, y, na_equal, info, n_row);
+  case vctrs_type_logical:   COMPARE_COL(int, LOGICAL_RO, lgl_compare_scalar); break;
+  case vctrs_type_integer:   COMPARE_COL(int, INTEGER_RO, int_compare_scalar); break;
+  case vctrs_type_double:    COMPARE_COL(double, REAL_RO, dbl_compare_scalar); break;
+  case vctrs_type_character: COMPARE_COL(SEXP, STRING_PTR_RO, chr_compare_scalar); break;
+  case vctrs_type_dataframe: df_compare_impl(p_out, p_info, x, y, na_equal); break;
   case vctrs_type_scalar:    Rf_errorcall(R_NilValue, "Can't compare scalars with `vctrs_compare()`");
   case vctrs_type_list:      Rf_errorcall(R_NilValue, "Can't compare lists with `vctrs_compare()`");
   default:                   Rf_error("Unimplemented type in `vctrs_compare()`");
