@@ -81,7 +81,7 @@ int equal_scalar_na_propagate_p(enum vctrs_type proxy_type,
 
 // -----------------------------------------------------------------------------
 
-static SEXP df_equal(SEXP x, SEXP y, bool na_equal, R_len_t n_row);
+static SEXP df_equal(SEXP x, SEXP y, bool na_equal, R_len_t size);
 
 #define EQUAL(CTYPE, CONST_DEREF, SCALAR_EQUAL)         \
   do {                                                  \
@@ -160,7 +160,7 @@ SEXP vctrs_equal(SEXP x, SEXP y, SEXP na_equal_) {
     const CTYPE* p_y = CONST_DEREF(y);                    \
                                                           \
     for (R_len_t i = 0; i < n; ++i, ++p_x, ++p_y) {       \
-      if (!SCALAR_EQUAL(p_x, p_y, true)) {                \
+      if (!SCALAR_EQUAL(p_x, p_y)) {                      \
         return false;                                     \
       }                                                   \
     }                                                     \
@@ -171,7 +171,7 @@ SEXP vctrs_equal(SEXP x, SEXP y, SEXP na_equal_) {
 #define EQUAL_ALL_BARRIER(SCALAR_EQUAL)                   \
   do {                                                    \
     for (R_len_t i = 0; i < n; ++i) {                     \
-      if (!SCALAR_EQUAL(x, i, y, i, true)) {              \
+      if (!SCALAR_EQUAL(x, i, y, i)) {                    \
         return false;                                     \
       }                                                   \
     }                                                     \
@@ -279,14 +279,14 @@ bool equal_object(SEXP x, SEXP y) {
   }
 
   switch (type) {
-  case LGLSXP:  EQUAL_ALL(int, LOGICAL_RO, lgl_equal_scalar);
-  case INTSXP:  EQUAL_ALL(int, INTEGER_RO, int_equal_scalar);
-  case REALSXP: EQUAL_ALL(double, REAL_RO, dbl_equal_scalar);
-  case STRSXP:  EQUAL_ALL(SEXP, STRING_PTR_RO, chr_equal_scalar);
-  case RAWSXP:  EQUAL_ALL(Rbyte, RAW_RO, raw_equal_scalar);
-  case CPLXSXP: EQUAL_ALL(Rcomplex, COMPLEX_RO, cpl_equal_scalar);
+  case LGLSXP:  EQUAL_ALL(int, LOGICAL_RO, lgl_equal_scalar_na_equal);
+  case INTSXP:  EQUAL_ALL(int, INTEGER_RO, int_equal_scalar_na_equal);
+  case REALSXP: EQUAL_ALL(double, REAL_RO, dbl_equal_scalar_na_equal);
+  case STRSXP:  EQUAL_ALL(SEXP, STRING_PTR_RO, chr_equal_scalar_na_equal);
+  case RAWSXP:  EQUAL_ALL(Rbyte, RAW_RO, raw_equal_scalar_na_equal);
+  case CPLXSXP: EQUAL_ALL(Rcomplex, COMPLEX_RO, cpl_equal_scalar_na_equal);
   case EXPRSXP:
-  case VECSXP:  EQUAL_ALL_BARRIER(list_equal_scalar);
+  case VECSXP:  EQUAL_ALL_BARRIER(list_equal_scalar_na_equal);
   default:      Rf_errorcall(R_NilValue, "Internal error: Unexpected type in `equal_object()`");
   }
 }
@@ -342,58 +342,17 @@ bool equal_names(SEXP x, SEXP y) {
 
 // -----------------------------------------------------------------------------
 
-static struct vctrs_df_rowwise_info vec_equal_col(SEXP x,
-                                                  SEXP y,
-                                                  bool na_equal,
-                                                  struct vctrs_df_rowwise_info info,
-                                                  R_len_t n_row);
+static void vec_equal_col(int* p_out,
+                          struct df_short_circuit_info* p_info,
+                          SEXP x,
+                          SEXP y,
+                          bool na_equal);
 
-static struct vctrs_df_rowwise_info df_equal_impl(SEXP x,
-                                                  SEXP y,
-                                                  bool na_equal,
-                                                  struct vctrs_df_rowwise_info info,
-                                                  R_len_t n_row);
-
-static struct vctrs_df_rowwise_info init_rowwise_equal_info(R_len_t n_row) {
-  struct vctrs_df_rowwise_info info;
-
-  // Initialize to "equality" value
-  // and only change if we learn that it differs
-  info.out = PROTECT(Rf_allocVector(LGLSXP, n_row));
-  int* p_out = LOGICAL(info.out);
-
-  for (R_len_t i = 0; i < n_row; ++i) {
-    p_out[i] = 1;
-  }
-
-  // To begin with, no rows have a known comparison value
-  info.row_known = PROTECT(Rf_allocVector(RAWSXP, n_row * sizeof(bool)));
-  info.p_row_known = (bool*) RAW(info.row_known);
-  memset(info.p_row_known, false, n_row * sizeof(bool));
-
-  info.remaining = n_row;
-
-  UNPROTECT(2);
-  return info;
-}
-
-static SEXP df_equal(SEXP x, SEXP y, bool na_equal, R_len_t n_row) {
-  int nprot = 0;
-
-  struct vctrs_df_rowwise_info info = init_rowwise_equal_info(n_row);
-  PROTECT_DF_ROWWISE_INFO(&info, &nprot);
-
-  info = df_equal_impl(x, y, na_equal, info, n_row);
-
-  UNPROTECT(nprot);
-  return info.out;
-}
-
-static struct vctrs_df_rowwise_info df_equal_impl(SEXP x,
-                                                  SEXP y,
-                                                  bool na_equal,
-                                                  struct vctrs_df_rowwise_info info,
-                                                  R_len_t n_row) {
+static void df_equal_impl(int* p_out,
+                          struct df_short_circuit_info* p_info,
+                          SEXP x,
+                          SEXP y,
+                          bool na_equal) {
   int n_col = Rf_length(x);
 
   if (n_col != Rf_length(y)) {
@@ -404,28 +363,46 @@ static struct vctrs_df_rowwise_info df_equal_impl(SEXP x,
     SEXP x_col = VECTOR_ELT(x, i);
     SEXP y_col = VECTOR_ELT(y, i);
 
-    info = vec_equal_col(x_col, y_col, na_equal, info, n_row);
+    vec_equal_col(p_out, p_info, x_col, y_col, na_equal);
 
     // If we know all comparison values, break
-    if (info.remaining == 0) {
+    if (p_info->remaining == 0) {
       break;
     }
   }
+}
 
-  return info;
+static SEXP df_equal(SEXP x, SEXP y, bool na_equal, R_len_t size) {
+  int nprot = 0;
+
+  SEXP out = PROTECT_N(Rf_allocVector(LGLSXP, size), &nprot);
+  int* p_out = LOGICAL(out);
+
+  // Initialize to "equality" value
+  // and only change if we learn that it differs
+  for (R_len_t i = 0; i < size; ++i) {
+    p_out[i] = 1;
+  }
+
+  struct df_short_circuit_info info = new_df_short_circuit_info(size);
+  struct df_short_circuit_info* p_info = &info;
+  PROTECT_DF_SHORT_CIRCUIT_INFO(p_info, &nprot);
+
+  df_equal_impl(p_out, p_info, x, y, na_equal);
+
+  UNPROTECT(nprot);
+  return out;
 }
 
 // -----------------------------------------------------------------------------
 
 #define EQUAL_COL(CTYPE, CONST_DEREF, SCALAR_EQUAL)                  \
 do {                                                                 \
-  int* p_out = LOGICAL(info.out);                                    \
-                                                                     \
   const CTYPE* p_x = CONST_DEREF(x);                                 \
   const CTYPE* p_y = CONST_DEREF(y);                                 \
                                                                      \
-  for (R_len_t i = 0; i < n_row; ++i, ++p_x, ++p_y) {                \
-    if (info.p_row_known[i]) {                                       \
+  for (R_len_t i = 0; i < p_info->size; ++i, ++p_x, ++p_y) {         \
+    if (p_info->p_row_known[i]) {                                    \
       continue;                                                      \
     }                                                                \
                                                                      \
@@ -433,25 +410,21 @@ do {                                                                 \
                                                                      \
     if (eq <= 0) {                                                   \
       p_out[i] = eq;                                                 \
-      info.p_row_known[i] = true;                                    \
-      --info.remaining;                                              \
+      p_info->p_row_known[i] = true;                                 \
+      --p_info->remaining;                                           \
                                                                      \
-      if (info.remaining == 0) {                                     \
+      if (p_info->remaining == 0) {                                  \
         break;                                                       \
       }                                                              \
     }                                                                \
   }                                                                  \
-                                                                     \
-  return info;                                                       \
 }                                                                    \
 while (0)
 
 #define EQUAL_COL_BARRIER(SCALAR_EQUAL)                \
 do {                                                   \
-  int* p_out = LOGICAL(info.out);                      \
-                                                       \
-  for (R_len_t i = 0; i < n_row; ++i) {                \
-    if (info.p_row_known[i]) {                         \
+  for (R_len_t i = 0; i < p_info->size; ++i) {         \
+    if (p_info->p_row_known[i]) {                      \
       continue;                                        \
     }                                                  \
                                                        \
@@ -459,33 +432,31 @@ do {                                                   \
                                                        \
     if (eq <= 0) {                                     \
       p_out[i] = eq;                                   \
-      info.p_row_known[i] = true;                      \
-      --info.remaining;                                \
+      p_info->p_row_known[i] = true;                   \
+      --p_info->remaining;                             \
                                                        \
-      if (info.remaining == 0) {                       \
+      if (p_info->remaining == 0) {                    \
         break;                                         \
       }                                                \
     }                                                  \
   }                                                    \
-                                                       \
-  return info;                                         \
 }                                                      \
 while (0)
 
-static struct vctrs_df_rowwise_info vec_equal_col(SEXP x,
-                                                  SEXP y,
-                                                  bool na_equal,
-                                                  struct vctrs_df_rowwise_info info,
-                                                  R_len_t n_row) {
+static void vec_equal_col(int* p_out,
+                          struct df_short_circuit_info* p_info,
+                          SEXP x,
+                          SEXP y,
+                          bool na_equal) {
   switch (vec_proxy_typeof(x)) {
-  case vctrs_type_logical:   EQUAL_COL(int, LOGICAL_RO, lgl_equal_scalar);
-  case vctrs_type_integer:   EQUAL_COL(int, INTEGER_RO, int_equal_scalar);
-  case vctrs_type_double:    EQUAL_COL(double, REAL_RO, dbl_equal_scalar);
-  case vctrs_type_raw:       EQUAL_COL(Rbyte, RAW_RO, raw_equal_scalar);
-  case vctrs_type_complex:   EQUAL_COL(Rcomplex, COMPLEX_RO, cpl_equal_scalar);
-  case vctrs_type_character: EQUAL_COL(SEXP, STRING_PTR_RO, chr_equal_scalar);
-  case vctrs_type_list:      EQUAL_COL_BARRIER(list_equal_scalar);
-  case vctrs_type_dataframe: return df_equal_impl(x, y, na_equal, info, n_row);
+  case vctrs_type_logical:   EQUAL_COL(int, LOGICAL_RO, lgl_equal_scalar); break;
+  case vctrs_type_integer:   EQUAL_COL(int, INTEGER_RO, int_equal_scalar); break;
+  case vctrs_type_double:    EQUAL_COL(double, REAL_RO, dbl_equal_scalar); break;
+  case vctrs_type_raw:       EQUAL_COL(Rbyte, RAW_RO, raw_equal_scalar); break;
+  case vctrs_type_complex:   EQUAL_COL(Rcomplex, COMPLEX_RO, cpl_equal_scalar); break;
+  case vctrs_type_character: EQUAL_COL(SEXP, STRING_PTR_RO, chr_equal_scalar); break;
+  case vctrs_type_list:      EQUAL_COL_BARRIER(list_equal_scalar); break;
+  case vctrs_type_dataframe: df_equal_impl(p_out, p_info, x, y, na_equal); break;
   case vctrs_type_scalar:    Rf_errorcall(R_NilValue, "Can't compare scalars with `vctrs_equal()`");
   default:                   Rf_error("Unimplemented type in `vctrs_equal()`");
   }
@@ -496,13 +467,14 @@ static struct vctrs_df_rowwise_info vec_equal_col(SEXP x,
 
 // -----------------------------------------------------------------------------
 
-static int lgl_equal_na_scalar(const int* x);
-static int int_equal_na_scalar(const int* x);
-static int dbl_equal_na_scalar(const double* x);
-static int cpl_equal_na_scalar(const Rcomplex* x);
-static int chr_equal_na_scalar(const SEXP* x);
-static int list_equal_na_scalar(SEXP x, R_len_t i);
-static int df_equal_na_scalar(SEXP x, R_len_t i);
+static inline int lgl_equal_na_scalar(const int* x);
+static inline int int_equal_na_scalar(const int* x);
+static inline int dbl_equal_na_scalar(const double* x);
+static inline int cpl_equal_na_scalar(const Rcomplex* x);
+static inline int raw_equal_na_scalar(const Rbyte* x);
+static inline int chr_equal_na_scalar(const SEXP* x);
+static inline int list_equal_na_scalar(SEXP x, R_len_t i);
+static inline int df_equal_na_scalar(SEXP x, R_len_t i);
 
 // If `x` is a data frame, it must have been recursively proxied
 // beforehand so we can safely use `TYPEOF(x)`
@@ -512,6 +484,7 @@ int equal_na(SEXP x, R_len_t i) {
   case INTSXP: return int_equal_na_scalar(INTEGER(x) + i);
   case REALSXP: return dbl_equal_na_scalar(REAL(x) + i);
   case CPLXSXP: return cpl_equal_na_scalar(COMPLEX(x) + i);
+  case RAWSXP: return raw_equal_na_scalar(RAW(x) + i);
   case STRSXP: return chr_equal_na_scalar(STRING_PTR(x) + i);
   default: break;
   }
@@ -527,79 +500,98 @@ int equal_na(SEXP x, R_len_t i) {
 
 #define EQUAL_NA(CTYPE, CONST_DEREF, SCALAR_EQUAL_NA)     \
   do {                                                    \
-    const CTYPE* xp = CONST_DEREF(x);                     \
+    SEXP out = PROTECT(Rf_allocVector(LGLSXP, size));     \
+    int* p_out = LOGICAL(out);                            \
                                                           \
-    for (R_len_t i = 0; i < n; ++i, ++xp) {               \
-      p[i] = SCALAR_EQUAL_NA(xp);                         \
+    const CTYPE* p_x = CONST_DEREF(x);                    \
+                                                          \
+    for (R_len_t i = 0; i < size; ++i, ++p_x) {           \
+      p_out[i] = SCALAR_EQUAL_NA(p_x);                    \
     }                                                     \
+                                                          \
+    UNPROTECT(2);                                         \
+    return out;                                           \
   }                                                       \
   while (0)
 
 #define EQUAL_NA_BARRIER(SCALAR_EQUAL_NA)                 \
   do {                                                    \
-    for (R_len_t i = 0; i < n; ++i) {                     \
-      p[i] = SCALAR_EQUAL_NA(x, i);                       \
+    SEXP out = PROTECT(Rf_allocVector(LGLSXP, size));     \
+    int* p_out = LOGICAL(out);                            \
+                                                          \
+    for (R_len_t i = 0; i < size; ++i) {                  \
+      p_out[i] = SCALAR_EQUAL_NA(x, i);                   \
     }                                                     \
+                                                          \
+    UNPROTECT(2);                                         \
+    return out;                                           \
   }                                                       \
   while (0)
 
+static SEXP df_equal_na(SEXP x, R_len_t size);
+
 // [[ register() ]]
 SEXP vctrs_equal_na(SEXP x) {
-  R_len_t n = vec_size(x);
-  SEXP out = PROTECT(Rf_allocVector(LGLSXP, n));
-  int32_t* p = LOGICAL(out);
+  R_len_t size = vec_size(x);
 
   x = PROTECT(vec_proxy_equal(x));
 
   enum vctrs_type type = vec_proxy_typeof(x);
 
   switch (type) {
-  case vctrs_type_logical:   EQUAL_NA(int, LOGICAL_RO, lgl_equal_na_scalar); break;
-  case vctrs_type_integer:   EQUAL_NA(int, INTEGER_RO, int_equal_na_scalar); break;
-  case vctrs_type_double:    EQUAL_NA(double, REAL_RO, dbl_equal_na_scalar); break;
-  case vctrs_type_complex:   EQUAL_NA(Rcomplex, COMPLEX_RO, cpl_equal_na_scalar); break;
-  case vctrs_type_character: EQUAL_NA(SEXP, STRING_PTR_RO, chr_equal_na_scalar); break;
-  case vctrs_type_list:      EQUAL_NA_BARRIER(list_equal_na_scalar); break;
-  case vctrs_type_dataframe: EQUAL_NA_BARRIER(df_equal_na_scalar); break;
+  case vctrs_type_logical:   EQUAL_NA(int, LOGICAL_RO, lgl_equal_na_scalar);
+  case vctrs_type_integer:   EQUAL_NA(int, INTEGER_RO, int_equal_na_scalar);
+  case vctrs_type_double:    EQUAL_NA(double, REAL_RO, dbl_equal_na_scalar);
+  case vctrs_type_complex:   EQUAL_NA(Rcomplex, COMPLEX_RO, cpl_equal_na_scalar);
+  case vctrs_type_raw:       EQUAL_NA(Rbyte, RAW_RO, raw_equal_na_scalar);
+  case vctrs_type_character: EQUAL_NA(SEXP, STRING_PTR_RO, chr_equal_na_scalar);
+  case vctrs_type_list:      EQUAL_NA_BARRIER(list_equal_na_scalar);
+  case vctrs_type_dataframe: {
+    SEXP out = df_equal_na(x, size);
+    UNPROTECT(1);
+    return out;
+  }
   case vctrs_type_scalar:    Rf_errorcall(R_NilValue, "Can't detect `NA` values in scalars with `vctrs_equal_na()`.");
   default:                   Rf_error("Unimplemented type in `vctrs_equal_na()`.");
   }
-
-  UNPROTECT(2);
-  return out;
 }
 
 #undef EQUAL_NA
 #undef EQUAL_NA_BARRIER
 
-static int lgl_equal_na_scalar(const int* x) {
+static inline int lgl_equal_na_scalar(const int* x) {
   return *x == NA_LOGICAL;
 }
 
-static int int_equal_na_scalar(const int* x) {
+static inline int int_equal_na_scalar(const int* x) {
   return *x == NA_INTEGER;
 }
 
-static int dbl_equal_na_scalar(const double* x) {
+static inline int dbl_equal_na_scalar(const double* x) {
   // is.na(NaN) is TRUE
   // isnan() does not consistently return 1 and 0 on all platforms,
   // but R's ISNAN() does
   return ISNAN(*x);
 }
 
-static int cpl_equal_na_scalar(const Rcomplex* x) {
+static inline int cpl_equal_na_scalar(const Rcomplex* x) {
   return ISNAN(x->r) || ISNAN(x->i);
 }
 
-static int chr_equal_na_scalar(const SEXP* x) {
+// Raw vectors can never be `NA`
+static inline int raw_equal_na_scalar(const Rbyte* x) {
+  return false;
+}
+
+static inline int chr_equal_na_scalar(const SEXP* x) {
   return *x == NA_STRING;
 }
 
-static int list_equal_na_scalar(SEXP x, R_len_t i) {
+static inline int list_equal_na_scalar(SEXP x, R_len_t i) {
   return Rf_isNull(VECTOR_ELT(x, i));
 }
 
-static int df_equal_na_scalar(SEXP x, R_len_t i) {
+static inline int df_equal_na_scalar(SEXP x, R_len_t i) {
   int n_col = Rf_length(x);
 
   for (int k = 0; k < n_col; ++k) {
@@ -610,3 +602,112 @@ static int df_equal_na_scalar(SEXP x, R_len_t i) {
 
   return true;
 }
+
+// -----------------------------------------------------------------------------
+
+static void vec_equal_na_col(int* p_out,
+                             struct df_short_circuit_info* p_info,
+                             SEXP x);
+
+static void df_equal_na_impl(int* p_out,
+                             struct df_short_circuit_info* p_info,
+                             SEXP x) {
+  int n_col = Rf_length(x);
+
+  for (R_len_t i = 0; i < n_col; ++i) {
+    SEXP col = VECTOR_ELT(x, i);
+
+    vec_equal_na_col(p_out, p_info, col);
+
+    // If all rows have at least one non-missing value, break
+    if (p_info->remaining == 0) {
+      break;
+    }
+  }
+}
+
+static SEXP df_equal_na(SEXP x, R_len_t size) {
+  int nprot = 0;
+
+  SEXP out = PROTECT_N(Rf_allocVector(LGLSXP, size), &nprot);
+  int* p_out = LOGICAL(out);
+
+  // Initialize to "equality" value
+  // and only change if we learn that it differs
+  for (R_len_t i = 0; i < size; ++i) {
+    p_out[i] = 1;
+  }
+
+  struct df_short_circuit_info info = new_df_short_circuit_info(size);
+  struct df_short_circuit_info* p_info = &info;
+  PROTECT_DF_SHORT_CIRCUIT_INFO(p_info, &nprot);
+
+  df_equal_na_impl(p_out, p_info, x);
+
+  UNPROTECT(nprot);
+  return out;
+}
+
+// -----------------------------------------------------------------------------
+
+#define EQUAL_NA_COL(CTYPE, CONST_DEREF, SCALAR_EQUAL_NA) \
+do {                                                      \
+  const CTYPE* p_x = CONST_DEREF(x);                      \
+                                                          \
+  for (R_len_t i = 0; i < p_info->size; ++i, ++p_x) {     \
+    if (p_info->p_row_known[i]) {                         \
+      continue;                                           \
+    }                                                     \
+                                                          \
+    if (!SCALAR_EQUAL_NA(p_x)) {                          \
+      p_out[i] = 0;                                       \
+      p_info->p_row_known[i] = true;                      \
+      --p_info->remaining;                                \
+                                                          \
+      if (p_info->remaining == 0) {                       \
+        break;                                            \
+      }                                                   \
+    }                                                     \
+  }                                                       \
+}                                                         \
+while (0)
+
+#define EQUAL_NA_COL_BARRIER(SCALAR_EQUAL_NA)  \
+do {                                           \
+  for (R_len_t i = 0; i < p_info->size; ++i) { \
+    if (p_info->p_row_known[i]) {              \
+      continue;                                \
+    }                                          \
+                                               \
+    if (!SCALAR_EQUAL_NA(x, i)) {              \
+      p_out[i] = 0;                            \
+      p_info->p_row_known[i] = true;           \
+      --p_info->remaining;                     \
+                                               \
+      if (p_info->remaining == 0) {            \
+        break;                                 \
+      }                                        \
+    }                                          \
+  }                                            \
+}                                              \
+while (0)
+
+static void vec_equal_na_col(int* p_out,
+                             struct df_short_circuit_info* p_info,
+                             SEXP x) {
+  switch (vec_proxy_typeof(x)) {
+  case vctrs_type_logical:   EQUAL_NA_COL(int, LOGICAL_RO, lgl_equal_na_scalar); break;
+  case vctrs_type_integer:   EQUAL_NA_COL(int, INTEGER_RO, int_equal_na_scalar); break;
+  case vctrs_type_double:    EQUAL_NA_COL(double, REAL_RO, dbl_equal_na_scalar); break;
+  case vctrs_type_complex:   EQUAL_NA_COL(Rcomplex, COMPLEX_RO, cpl_equal_na_scalar); break;
+  case vctrs_type_raw:       EQUAL_NA_COL(Rbyte, RAW_RO, raw_equal_na_scalar); break;
+  case vctrs_type_character: EQUAL_NA_COL(SEXP, STRING_PTR_RO, chr_equal_na_scalar); break;
+  case vctrs_type_list:      EQUAL_NA_COL_BARRIER(list_equal_na_scalar); break;
+  case vctrs_type_dataframe: df_equal_na_impl(p_out, p_info, x); break;
+  case vctrs_type_scalar:    Rf_errorcall(R_NilValue, "Can't compare scalars with `vec_equal_na()`");
+  default:                   Rf_error("Unimplemented type in `vec_equal_na()`");
+  }
+}
+
+#undef EQUAL_NA_COL
+#undef EQUAL_NA_COL_BARRIER
