@@ -4,32 +4,32 @@
 #include "subscript-loc.h"
 
 static SEXP int_invert_location(SEXP subscript, R_len_t n,
-                                const struct vec_as_location_opts* opts);
+                                const struct location_opts* opts);
 static SEXP int_filter_zero(SEXP subscript, R_len_t n_zero);
 static void int_check_consecutive(SEXP subscript, R_len_t n,
-                                  const struct vec_as_location_opts* opts);
+                                  const struct location_opts* opts);
 
 static void stop_subscript_missing(SEXP i);
 static void stop_subscript_oob_location(SEXP i, R_len_t size,
-                                        const struct vec_as_location_opts* opts);
+                                        const struct location_opts* opts);
 static void stop_subscript_oob_name(SEXP i, SEXP names,
-                                    const struct vec_as_location_opts* opts);
+                                    const struct location_opts* opts);
 static void stop_location_negative(SEXP i,
-                                   const struct vec_as_location_opts* opts);
+                                   const struct location_opts* opts);
 static void stop_location_zero(SEXP i,
-                               const struct vec_as_location_opts* opts);
+                               const struct location_opts* opts);
 static void stop_indicator_size(SEXP i, SEXP n,
-                                const struct vec_as_location_opts* opts);
+                                const struct location_opts* opts);
 static void stop_location_negative_missing(SEXP i,
-                                           const struct vec_as_location_opts* opts);
+                                           const struct location_opts* opts);
 static void stop_location_negative_positive(SEXP i,
-                                            const struct vec_as_location_opts* opts);
+                                            const struct location_opts* opts);
 static void stop_location_oob_non_consecutive(SEXP i, R_len_t size,
-                                              const struct vec_as_location_opts* opts);
+                                              const struct location_opts* opts);
 
 
 static SEXP int_as_location(SEXP subscript, R_len_t n,
-                            const struct vec_as_location_opts* opts) {
+                            const struct location_opts* opts) {
   const int* data = INTEGER_RO(subscript);
   R_len_t loc_n = Rf_length(subscript);
 
@@ -86,10 +86,10 @@ static SEXP int_as_location(SEXP subscript, R_len_t n,
 
 
 static SEXP lgl_as_location(SEXP subscript, R_len_t n,
-                            const struct vec_as_location_opts* opts);
+                            const struct location_opts* opts);
 
 static SEXP int_invert_location(SEXP subscript, R_len_t n,
-                                const struct vec_as_location_opts* opts) {
+                                const struct location_opts* opts) {
   const int* data = INTEGER_RO(subscript);
   R_len_t loc_n = Rf_length(subscript);
 
@@ -114,8 +114,10 @@ static SEXP int_invert_location(SEXP subscript, R_len_t n,
 
     j = -j;
     if (j > n) {
-      struct vec_as_location_opts updated_opts = *opts;
-      updated_opts.action = SUBSCRIPT_ACTION_NEGATE;
+      struct location_opts updated_opts = *opts;
+      struct subscript_opts updated_subscript_opts = *updated_opts.subscript_opts;
+      updated_subscript_opts.action = SUBSCRIPT_ACTION_NEGATE;
+      updated_opts.subscript_opts = &updated_subscript_opts;
       stop_subscript_oob_location(subscript, n, &updated_opts);
     }
 
@@ -151,7 +153,7 @@ static SEXP int_filter_zero(SEXP subscript, R_len_t n_zero) {
 int qsort_icmp(const void* x, const void* y);
 
 static void int_check_consecutive(SEXP subscript, R_len_t n,
-                                  const struct vec_as_location_opts* opts) {
+                                  const struct location_opts* opts) {
   SEXP sorted = PROTECT(Rf_duplicate(subscript));
   int* p_sorted = INTEGER(sorted);
 
@@ -182,7 +184,7 @@ static void int_check_consecutive(SEXP subscript, R_len_t n,
 }
 
 static SEXP dbl_as_location(SEXP subscript, R_len_t n,
-                            const struct vec_as_location_opts* opts) {
+                            const struct location_opts* opts) {
   subscript = PROTECT(vec_cast(subscript, vctrs_shared_empty_int, args_empty, args_empty));
   subscript = int_as_location(subscript, n, opts);
 
@@ -191,7 +193,7 @@ static SEXP dbl_as_location(SEXP subscript, R_len_t n,
 }
 
 static SEXP lgl_as_location(SEXP subscript, R_len_t n,
-                            const struct vec_as_location_opts* opts) {
+                            const struct location_opts* opts) {
   R_len_t subscript_n = Rf_length(subscript);
 
   if (subscript_n == n) {
@@ -250,7 +252,7 @@ static SEXP lgl_as_location(SEXP subscript, R_len_t n,
 }
 
 static SEXP chr_as_location(SEXP subscript, SEXP names,
-                            const struct vec_as_location_opts* opts) {
+                            const struct location_opts* opts) {
   if (names == R_NilValue) {
     Rf_errorcall(R_NilValue, "Can't use character names to index an unnamed vector.");
   }
@@ -280,30 +282,14 @@ SEXP vec_as_location(SEXP subscript, R_len_t n, SEXP names) {
   return vec_as_location_opts(subscript,
                               n,
                               names,
-                              vec_as_location_default_opts,
-                              NULL);
+                              location_default_opts);
 }
 
 SEXP vec_as_location_opts(SEXP subscript, R_len_t n, SEXP names,
-                          const struct vec_as_location_opts* location_opts,
-                          const struct vec_as_subscript_opts* subscript_opts) {
-
-  if (vec_dim_n(subscript) != 1) {
-    Rf_errorcall(R_NilValue, "`i` must have one dimension, not %d.", vec_dim_n(subscript));
-  }
+                          const struct location_opts* opts) {
 
   ERR err = NULL;
-  if (subscript_opts) {
-    subscript = vec_as_subscript_opts(subscript, subscript_opts, &err);
-  } else {
-    struct vec_as_subscript_opts default_subscript_opts = {
-      .logical = SUBSCRIPT_TYPE_ACTION_CAST,
-      .numeric = SUBSCRIPT_TYPE_ACTION_CAST,
-      .character = SUBSCRIPT_TYPE_ACTION_CAST,
-      .subscript_arg = location_opts->subscript_arg
-    };
-    subscript = vec_as_subscript_opts(subscript, &default_subscript_opts, &err);
-  }
+  subscript = vec_as_subscript_opts(subscript, opts->subscript_opts, &err);
   PROTECT2(subscript, err);
 
   if (err) {
@@ -314,11 +300,11 @@ SEXP vec_as_location_opts(SEXP subscript, R_len_t n, SEXP names,
   SEXP out = R_NilValue;
   switch (TYPEOF(subscript)) {
   case NILSXP: out = vctrs_shared_empty_int; break;
-  case INTSXP: out = int_as_location(subscript, n, location_opts); break;
-  case REALSXP: out = dbl_as_location(subscript, n, location_opts); break;
-  case LGLSXP: out = lgl_as_location(subscript, n, location_opts); break;
-  case STRSXP: out = chr_as_location(subscript, names, location_opts); break;
-  default: Rf_errorcall(R_NilValue, "`i` must be an integer, character, or logical vector, not a %s.",
+  case INTSXP: out = int_as_location(subscript, n, opts); break;
+  case REALSXP: out = dbl_as_location(subscript, n, opts); break;
+  case LGLSXP: out = lgl_as_location(subscript, n, opts); break;
+  case STRSXP: out = chr_as_location(subscript, names, opts); break;
+  default: Rf_errorcall(R_NilValue, "Internal error: Wrong subscript type `%s` in `vec_as_location_opts()`.",
                         Rf_type2char(TYPEOF(subscript)));
   }
 
@@ -352,7 +338,7 @@ static enum subscript_missing parse_subscript_arg_missing(SEXP x) {
 
   never_reached("stop_subscript_arg_missing");
 }
-static enum num_as_location_loc_negative parse_loc_negative(SEXP x) {
+static enum num_loc_negative parse_loc_negative(SEXP x) {
   if (TYPEOF(x) != STRSXP || Rf_length(x) == 0) {
     stop_bad_negative();
   }
@@ -366,7 +352,7 @@ static enum num_as_location_loc_negative parse_loc_negative(SEXP x) {
 
   never_reached("stop_bad_negative");
 }
-static enum num_as_location_loc_oob parse_loc_oob(SEXP x) {
+static enum num_loc_oob parse_loc_oob(SEXP x) {
   if (TYPEOF(x) != STRSXP || Rf_length(x) == 0) {
     stop_bad_oob();
   }
@@ -379,7 +365,7 @@ static enum num_as_location_loc_oob parse_loc_oob(SEXP x) {
 
   never_reached("stop_bad_oob");
 }
-static enum num_as_location_loc_zero parse_loc_zero(SEXP x) {
+static enum num_loc_zero parse_loc_zero(SEXP x) {
   if (TYPEOF(x) != STRSXP || Rf_length(x) == 0) {
     stop_bad_zero();
   }
@@ -404,7 +390,7 @@ SEXP vctrs_as_location(SEXP subscript, SEXP n_, SEXP names,
     n = Rf_length(subscript);
   } else {
     if (OBJECT(n_) || TYPEOF(n_) != INTSXP) {
-      n_ = vec_coercible_cast(n_, vctrs_shared_empty_int, args_empty, args_empty);
+      n_ = vec_cast(n_, vctrs_shared_empty_int, args_empty, args_empty);
     }
     PROTECT(n_);
 
@@ -417,16 +403,19 @@ SEXP vctrs_as_location(SEXP subscript, SEXP n_, SEXP names,
   }
 
   struct vctrs_arg arg = vec_as_arg(arg_);
-  struct vec_as_location_opts opts = {
-    .action = SUBSCRIPT_ACTION_DEFAULT,
-    .missing = parse_subscript_arg_missing(missing),
-    .loc_negative = parse_loc_negative(loc_negative),
-    .loc_oob = parse_loc_oob(loc_oob),
-    .loc_zero = parse_loc_zero(loc_zero),
-    .subscript_arg = &arg
+
+  struct subscript_opts subscript_opts = {
+    .subscript_arg  = &arg
+  };
+  struct location_opts opts = {
+    .subscript_opts = &subscript_opts,
+    .missing        = parse_subscript_arg_missing(missing),
+    .loc_negative   = parse_loc_negative(loc_negative),
+    .loc_oob        = parse_loc_oob(loc_oob),
+    .loc_zero       = parse_loc_zero(loc_zero)
   };
 
-  return vec_as_location_opts(subscript, n, names, &opts, NULL);
+  return vec_as_location_opts(subscript, n, names, &opts);
 }
 
 static void stop_subscript_missing(SEXP i) {
@@ -437,35 +426,35 @@ static void stop_subscript_missing(SEXP i) {
 }
 
 static void stop_location_negative_missing(SEXP i,
-                                           const struct vec_as_location_opts* opts) {
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+                                           const struct location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_negative_missing"),
                    syms_i, i,
                    syms_subscript_arg, arg,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    vctrs_ns_env);
   never_reached("stop_location_negative_missing");
 }
 static void stop_location_negative_positive(SEXP i,
-                                            const struct vec_as_location_opts* opts) {
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+                                            const struct location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_negative_positive"),
                    syms_i, i,
                    syms_subscript_arg, arg,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    vctrs_ns_env);
   never_reached("stop_location_negative_positive");
 }
 
 static void stop_subscript_oob_location(SEXP i, R_len_t size,
-                                        const struct vec_as_location_opts* opts) {
+                                        const struct location_opts* opts) {
   SEXP size_obj = PROTECT(r_int(size));
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask5(Rf_install("stop_subscript_oob"),
                    syms_i, i,
                    syms_subscript_type, chrs_numeric,
                    syms_size, size_obj,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    syms_subscript_arg, arg,
                    vctrs_ns_env);
 
@@ -473,60 +462,60 @@ static void stop_subscript_oob_location(SEXP i, R_len_t size,
   never_reached("stop_subscript_oob_location");
 }
 static void stop_subscript_oob_name(SEXP i, SEXP names,
-                                    const struct vec_as_location_opts* opts) {
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+                                    const struct location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask5(Rf_install("stop_subscript_oob"),
                    syms_i, i,
                    syms_subscript_type, chrs_character,
                    syms_names, names,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_subscript_oob_name");
 }
 
 static void stop_location_negative(SEXP i,
-                                   const struct vec_as_location_opts* opts) {
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+                                   const struct location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_negative"),
                    syms_i, i,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_location_negative");
 }
 
 static void stop_location_zero(SEXP i,
-                               const struct vec_as_location_opts* opts) {
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+                               const struct location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_zero"),
                    syms_i, i,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_location_zero");
 }
 
 static void stop_indicator_size(SEXP i, SEXP n,
-                                const struct vec_as_location_opts* opts) {
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+                                const struct location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask4(Rf_install("stop_indicator_size"),
                    syms_i, i,
                    syms_n, n,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_indicator_size");
 }
 
 static void stop_location_oob_non_consecutive(SEXP i, R_len_t size,
-                                              const struct vec_as_location_opts* opts) {
+                                              const struct location_opts* opts) {
   SEXP size_obj = PROTECT(r_int(size));
-  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_opts->subscript_arg));
   vctrs_eval_mask4(Rf_install("stop_location_oob_non_consecutive"),
                    syms_i, i,
                    syms_size, size_obj,
-                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_action, get_opts_action(opts->subscript_opts),
                    syms_subscript_arg, arg,
                    vctrs_ns_env);
 
@@ -534,22 +523,16 @@ static void stop_location_oob_non_consecutive(SEXP i, R_len_t size,
   never_reached("stop_location_oob_non_consecutive");
 }
 
-
-struct vec_as_location_opts vec_as_location_default_opts_obj;
-struct vec_as_location_opts vec_as_location_default_assign_opts_obj;
+struct location_opts location_default_opts_obj;
+struct location_opts location_default_assign_opts_obj;
 
 void vctrs_init_subscript_loc(SEXP ns) {
-  vec_as_location_default_opts_obj.action = SUBSCRIPT_ACTION_DEFAULT;
-  vec_as_location_default_opts_obj.loc_negative = LOC_NEGATIVE_INVERT;
-  vec_as_location_default_opts_obj.loc_oob = LOC_OOB_ERROR;
-  vec_as_location_default_opts_obj.loc_zero = LOC_ZERO_REMOVE;
-  vec_as_location_default_opts_obj.subscript_arg = NULL;
-  vec_as_location_default_opts_obj.missing = SUBSCRIPT_MISSING_PROPAGATE;
+  location_default_opts_obj.subscript_opts = &subscript_default_opts;
+  location_default_opts_obj.loc_negative = LOC_NEGATIVE_INVERT;
+  location_default_opts_obj.loc_oob = LOC_OOB_ERROR;
+  location_default_opts_obj.loc_zero = LOC_ZERO_REMOVE;
+  location_default_opts_obj.missing = SUBSCRIPT_MISSING_PROPAGATE;
 
-  vec_as_location_default_assign_opts_obj.action = SUBSCRIPT_ACTION_ASSIGN;
-  vec_as_location_default_assign_opts_obj.loc_negative = LOC_NEGATIVE_INVERT;
-  vec_as_location_default_assign_opts_obj.loc_oob = LOC_OOB_ERROR;
-  vec_as_location_default_assign_opts_obj.loc_zero = LOC_ZERO_REMOVE;
-  vec_as_location_default_assign_opts_obj.subscript_arg = NULL;
-  vec_as_location_default_assign_opts_obj.missing = SUBSCRIPT_MISSING_PROPAGATE;
+  location_default_assign_opts_obj = location_default_opts_obj;
+  location_default_assign_opts_obj.subscript_opts = &subscript_default_assign_opts;
 }
