@@ -11,10 +11,13 @@ test_that("vec_as_location2() requires integer or character inputs", {
     expect_error(vec_as_location2(env(), 10L), class = "vctrs_error_subscript_type")
     expect_error(vec_as_location2(foobar(), 10L), class = "vctrs_error_subscript_type")
     expect_error(vec_as_location2(2.5, 10L), class = "vctrs_error_subscript_type")
+    expect_error(vec_as_location2(Inf, 10L), class = "vctrs_error_subscript_type")
+    expect_error(vec_as_location2(-Inf, 10L), class = "vctrs_error_subscript_type")
 
     "Idem with custom `arg`"
     expect_error(vec_as_location2(foobar(), 10L, arg = "foo"), class = "vctrs_error_subscript_type")
     expect_error(vec_as_location2(2.5, 3L, arg = "foo"), class = "vctrs_error_subscript_type")
+    expect_error(with_tibble_rows(vec_as_location2(TRUE)), class = "vctrs_error_subscript_type")
   })
 })
 
@@ -26,6 +29,7 @@ test_that("vec_as_location() requires integer, character, or logical inputs", {
     expect_error(vec_as_location(2.5, 10L), class = "vctrs_error_subscript_type")
     expect_error(vec_as_location(list(), 10L), class = "vctrs_error_subscript_type")
     expect_error(vec_as_location(function() NULL, 10L), class = "vctrs_error_subscript_type")
+    expect_error(vec_as_location(Sys.Date(), 3L), class = "vctrs_error_subscript_type")
 
     "Idem with custom `arg`"
     expect_error(vec_as_location(env(), 10L, arg = "foo"), class = "vctrs_error_subscript_type")
@@ -42,8 +46,7 @@ test_that("vec_as_location2() and vec_as_location() require integer- or characte
 
   # Define subtype of logical and integer
   local_methods(
-    vec_ptype2.vctrs_foobar = function(x, y, ...) UseMethod("vec_ptype2.vctrs_foobar", y),
-    vec_ptype2.vctrs_foobar.default = function(x, y, ...) vec_default_ptype2(x, y, ...),
+    vec_ptype2.vctrs_foobar = function(x, y, ...) UseMethod("vec_ptype2.vctrs_foobar"),
     vec_ptype2.vctrs_foobar.logical = function(x, y, ...) logical(),
     vec_ptype2.vctrs_foobar.integer = function(x, y, ...) integer(),
     vec_ptype2.logical.vctrs_foobar = function(x, y, ...) logical(),
@@ -426,6 +429,18 @@ test_that("num_as_location() requires non-S3 inputs", {
   expect_error(num_as_location(factor("foo"), 2), "must be a numeric vector")
 })
 
+test_that("vec_as_location() checks dimensionality", {
+  verify_errors({
+    expect_error(vec_as_location(matrix(TRUE, nrow = 1), 3L), class = "vctrs_error_subscript_type")
+    expect_error(vec_as_location(array(TRUE, dim = c(1, 1, 1)), 3L), class = "vctrs_error_subscript_type")
+    expect_error(with_tibble_rows(vec_as_location(matrix(TRUE, nrow = 1), 3L)), class = "vctrs_error_subscript_type")
+  })
+})
+
+test_that("vec_as_location() works with vectors of dimensionality 1", {
+  expect_identical(vec_as_location(array(TRUE, dim = 1), 3L), 1:3)
+})
+
 test_that("conversion to locations has informative error messages", {
   verify_output(test_path("error", "test-subscript-loc.txt"), {
     "# vec_as_location() checks for mix of negative and missing locations"
@@ -456,6 +471,7 @@ test_that("conversion to locations has informative error messages", {
     vec_as_location(2.5, 3L)
     vec_as_location(list(), 10L)
     vec_as_location(function() NULL, 10L)
+    vec_as_location(Sys.Date(), 3L)
     "Idem with custom `arg`"
     vec_as_location(env(), 10L, arg = "foo")
     vec_as_location(foobar(), 10L, arg = "foo")
@@ -467,9 +483,12 @@ test_that("conversion to locations has informative error messages", {
     vec_as_location2(env(), 10L)
     vec_as_location2(foobar(), 10L)
     vec_as_location2(2.5, 3L)
+    vec_as_location2(Inf, 10L)
+    vec_as_location2(-Inf, 10L)
     "Idem with custom `arg`"
     vec_as_location2(foobar(), 10L, arg = "foo")
     vec_as_location2(2.5, 3L, arg = "foo")
+    with_tibble_rows(vec_as_location2(TRUE))
 
     "# vec_as_location2() requires length 1 inputs"
     vec_as_location2(1:2, 2L)
@@ -552,5 +571,10 @@ test_that("conversion to locations has informative error messages", {
     vec_as_location(c(1, NA), 2, missing = "error")
     vec_as_location(c(1, NA, 2, NA), 2, missing = "error", arg = "foo")
     with_tibble_cols(vec_as_location(c(1, NA, 2, NA), 2, missing = "error"))
+
+    "# vec_as_location() checks dimensionality"
+    vec_as_location(matrix(TRUE, nrow = 1), 3L)
+    vec_as_location(array(TRUE, dim = c(1, 1, 1)), 3L)
+    with_tibble_rows(vec_as_location(matrix(TRUE, nrow = 1), 3L))
   })
 })
